@@ -14,14 +14,14 @@ use Scalar::Util 'weaken';
 use constant { POE_IO_READ => 0, POE_IO_WRITE => 1 };
 use constant DEBUG => $ENV{MOJO_REACTOR_POE_DEBUG} || 0;
 
-our $VERSION = '0.007';
+our $VERSION = '0.008';
 
 my $POE;
 
 sub DESTROY {
 	my $self = shift;
 	$self->reset;
-	$self->_send_clear_all;
+	$self->_session_call_if_exists('mojo_clear_all');
 	undef $POE;
 }
 
@@ -29,7 +29,8 @@ sub again {
 	my ($self, $id) = @_;
 	croak 'Timer not active' unless my $timer = $self->{timers}{$id};
 	$timer->{time} = steady_time + $timer->{after};
-	$self->_send_adjust_timer($id);
+	# If session doesn't exist, the time will be set when it starts
+	$self->_session_call_if_exists(mojo_adjust_timer => $id);
 }
 
 sub io {
@@ -46,6 +47,13 @@ sub is_running {
 # We have to fall back to Mojo::Reactor::Poll, since POE::Kernel is unique
 sub new { $POE++ ? Mojo::Reactor::Poll->new : shift->SUPER::new }
 
+sub next_tick {
+	my ($self, $cb) = @_;
+	push @{$self->{next_tick}}, $cb;
+	$self->{next_timer} //= $self->timer(0 => \&_next);
+	return undef;
+}
+
 sub one_tick { shift->_init_session; POE::Kernel->run_one_timeslice; }
 
 sub recurring { shift->_timer(1, @_) }
@@ -56,13 +64,15 @@ sub remove {
 	if (ref $remove) {
 		if (exists $self->{io}{fileno $remove}) {
 			warn "-- Removed IO watcher for ".fileno($remove)."\n" if DEBUG;
-			$self->_send_clear_io(fileno $remove);
+			# If session doesn't exist, the watcher won't be re-added
+			$self->_session_call_if_exists(mojo_clear_io => fileno $remove);
 		}
 		return !!delete $self->{io}{fileno $remove};
 	} else {
 		if (exists $self->{timers}{$remove}) {
 			warn "-- Removed timer $remove\n" if DEBUG;
-			$self->_send_clear_timer($remove);
+			# If session doesn't exist, the timer won't be re-added
+			$self->_session_call_if_exists(mojo_clear_timer => $remove);
 		}
 		return !!delete $self->{timers}{$remove};
 	}
@@ -72,6 +82,7 @@ sub reset {
 	my $self = shift;
 	$self->remove($_) for keys %{$self->{timers}};
 	$self->remove($self->{io}{$_}{handle}) for keys %{$self->{io}};
+	delete @{$self}{qw(next_tick next_timer)};
 }
 
 sub start { shift->_init_session; POE::Kernel->run; }
@@ -90,7 +101,7 @@ sub watch {
 	
 	warn "-- Set IO watcher for ".fileno($handle)."\n" if DEBUG;
 	
-	$self->_send_set_io(fileno $handle);
+	$self->_session_call(mojo_set_io => fileno $handle);
 	
 	return $self;
 }
@@ -102,9 +113,10 @@ sub _id {
 	return $id;
 }
 
-sub _sandbox {
-	my ($self, $event, $cb) = (shift, shift, shift);
-	eval { $self->$cb(@_); 1 } or $self->emit(error => "$event failed: $@");
+sub _next {
+	my $self = shift;
+	delete $self->{next_timer};
+	while (my $cb = shift @{$self->{next_tick}}) { $self->$cb }
 }
 
 sub _timer {
@@ -120,35 +132,32 @@ sub _timer {
 		warn "-- Set timer $id after $after seconds$is_recurring\n";
 	}
 	
-	$self->_send_set_timer($id);
+	$self->_session_call(mojo_set_timer => $id);
 	
 	return $id;
 }
 
-sub _session_exists {
-	my $self = shift;
-	return undef unless defined $self->{session_id};
-	return !!POE::Kernel->ID_id_to_session($self->{session_id});
+sub _try {
+	my ($self, $what, $cb) = (shift, shift, shift);
+	eval { $self->$cb(@_); 1 } or $self->emit(error => "$what failed: $@");
 }
 
 sub _init_session {
 	my $self = shift;
 	unless ($self->_session_exists) {
 		my $session = POE::Session->create(
-			package_states => [
-				__PACKAGE__, {
-					_start				=> '_event_start',
-					_stop				=> '_event_stop',
-					mojo_set_timer 		=> '_event_set_timer',
-					mojo_clear_timer	=> '_event_clear_timer',
-					mojo_adjust_timer	=> '_event_adjust_timer',
-					mojo_clear_all		=> '_event_clear_all',
-					mojo_set_io			=> '_event_set_io',
-					mojo_clear_io		=> '_event_clear_io',
-					mojo_timer			=> '_event_timer',
-					mojo_io				=> '_event_io',
-				},
-			],
+			inline_states => {
+				_start				=> \&_event_start,
+				_stop				=> \&_event_stop,
+				mojo_set_timer 		=> \&_event_set_timer,
+				mojo_clear_timer	=> \&_event_clear_timer,
+				mojo_adjust_timer	=> \&_event_adjust_timer,
+				mojo_clear_all		=> \&_event_clear_all,
+				mojo_set_io			=> \&_event_set_io,
+				mojo_clear_io		=> \&_event_clear_io,
+				mojo_timer			=> \&_event_timer,
+				mojo_io				=> \&_event_io,
+			},
 			heap => { mojo_reactor => $self },
 		);
 		weaken $session->get_heap()->{mojo_reactor};
@@ -157,49 +166,27 @@ sub _init_session {
 	return $self;
 }
 
-sub _send_adjust_timer {
-	my ($self, $id) = @_;
-	# If session doesn't exist, the time will be set when it starts
-	return unless $self->_session_exists;
-	POE::Kernel->call($self->{session_id}, mojo_adjust_timer => $id);
-}
-
-sub _send_set_timer {
-	my ($self, $id) = @_;
-	# We need a session to set a timer
-	$self->_init_session;
-	POE::Kernel->call($self->{session_id}, mojo_set_timer => $id);
-}
-
-sub _send_clear_timer {
-	my ($self, $id) = @_;
-	# If session doesn't exist, the timer won't be re-added
-	return unless $self->_session_exists;
-	POE::Kernel->call($self->{session_id}, mojo_clear_timer => $id);
-}
-
-sub _send_clear_all {
+sub _session_exists {
 	my $self = shift;
-	return unless $self->_session_exists;
-	POE::Kernel->call($self->{session_id}, 'mojo_clear_all')
+	return undef unless defined $self->{session_id};
+	return !!POE::Kernel->ID_id_to_session($self->{session_id});
 }
 
-sub _send_set_io {
-	my ($self, $fd) = @_;
-	# We need a session to set a watcher
+sub _session_call_if_exists {
+	my $self = shift;
+	POE::Kernel->call($self->{session_id}, @_) if $self->_session_exists;
+	return $self;
+}
+
+sub _session_call {
+	my $self = shift;
 	$self->_init_session;
-	POE::Kernel->call($self->{session_id}, mojo_set_io => $fd);
-}
-
-sub _send_clear_io {
-	my ($self, $fd) = @_;
-	# If session doesn't exist, the watcher won't be re-added
-	return unless $self->_session_exists;
-	POE::Kernel->call($self->{session_id}, mojo_clear_io => $fd);
+	POE::Kernel->call($self->{session_id}, @_);
+	return $self;
 }
 
 sub _event_start {
-	my $self = $_[HEAP]->{mojo_reactor};
+	my $self = $_[HEAP]{mojo_reactor};
 	my $session = $_[SESSION];
 	
 	warn "-- POE session started\n" if DEBUG;
@@ -215,11 +202,11 @@ sub _event_start {
 			POE::Kernel->call($session, mojo_set_io => $fd);
 		}
 	}
-	undef $self->{queue};
+	delete $self->{queue};
 }
 
 sub _event_stop {
-	my $self = $_[HEAP]->{mojo_reactor};
+	my $self = $_[HEAP]{mojo_reactor};
 	
 	warn "-- POE session stopped\n" if DEBUG;
 	
@@ -232,12 +219,12 @@ sub _event_stop {
 		io => [keys %{$self->{io}}]
 	};
 	
-	undef $self->{session_id};
+	delete $self->{session_id};
 }
 
 sub _event_set_timer {
-	my ($heap, $id) = @_[HEAP, ARG0];
-	my $self = $heap->{mojo_reactor};
+	my $self = $_[HEAP]{mojo_reactor};
+	my $id = $_[ARG0];
 	return unless exists $self->{timers}{$id}
 		and defined $self->{timers}{$id}{time};
 	my $timer = $self->{timers}{$id};
@@ -249,8 +236,8 @@ sub _event_set_timer {
 }
 
 sub _event_clear_timer {
-	my ($heap, $id) = @_[HEAP, ARG0];
-	my $self = $heap->{mojo_reactor};
+	my $self = $_[HEAP]{mojo_reactor};
+	my $id = $_[ARG0];
 	return unless exists $self->{timers}{$id}
 		and defined $self->{timers}{$id}{poe_id};
 	my $timer = $self->{timers}{$id};
@@ -260,8 +247,8 @@ sub _event_clear_timer {
 }
 
 sub _event_adjust_timer {
-	my ($heap, $id) = @_[HEAP, ARG0];
-	my $self = $heap->{mojo_reactor};
+	my $self = $_[HEAP]{mojo_reactor};
+	my $id = $_[ARG0];
 	return unless exists $self->{timers}{$id}
 		and defined $self->{timers}{$id}{time}
 		and defined $self->{timers}{$id}{poe_id};
@@ -274,15 +261,15 @@ sub _event_adjust_timer {
 }
 
 sub _event_clear_all {
-	my $self = $_[HEAP]->{mojo_reactor} // return;
+	my $self = $_[HEAP]{mojo_reactor} // return;
 	POE::Kernel->alarm_remove_all;
 	
 	warn "-- Cleared all POE timers\n" if DEBUG;
 }
 
 sub _event_set_io {
-	my ($heap, $fd) = @_[HEAP, ARG0];
-	my $self = $heap->{mojo_reactor};
+	my $self = $_[HEAP]{mojo_reactor};
+	my $fd = $_[ARG0];
 	return unless exists $self->{io}{$fd}
 		and defined $self->{io}{$fd}{handle};
 	my $io = $self->{io}{$fd};
@@ -302,8 +289,8 @@ sub _event_set_io {
 }
 
 sub _event_clear_io {
-	my ($heap, $fd) = @_[HEAP, ARG0];
-	my $self = $heap->{mojo_reactor};
+	my $self = $_[HEAP]{mojo_reactor};
+	my $fd = $_[ARG0];
 	return unless exists $self->{io}{$fd}
 		and defined $self->{io}{$fd}{handle};
 	my $io = $self->{io}{$fd};
@@ -315,31 +302,31 @@ sub _event_clear_io {
 }
 
 sub _event_timer {
-	my ($heap, $id) = @_[HEAP,ARG0];
-	my $self = $heap->{mojo_reactor};
+	my $self = $_[HEAP]{mojo_reactor};
+	my $id = $_[ARG0];
 	
 	my $timer = $self->{timers}{$id};
 	warn "-- Event fired for timer $id\n" if DEBUG;
 	if (exists $timer->{recurring}) {
 		$timer->{time} = steady_time + $timer->{recurring};
-		$self->_send_set_timer($id);
+		$self->_session_call(mojo_set_timer => $id);
 	} else {
 		delete $self->{timers}{$id};
 	}
 	
-	$self->_sandbox("Timer $id", $timer->{cb});
+	$self->_try('Timer', $timer->{cb});
 }
 
 sub _event_io {
-	my ($heap, $handle, $mode) = @_[HEAP, ARG0, ARG1];
-	my $self = $heap->{mojo_reactor};
+	my $self = $_[HEAP]{mojo_reactor};
+	my ($handle, $mode) = @_[ARG0, ARG1];
 	
 	my $io = $self->{io}{fileno $handle};
 	#warn "-- Event fired for IO watcher ".fileno($handle)."\n" if DEBUG;
 	if ($mode == POE_IO_READ) {
-		$self->_sandbox('Read', $io->{cb}, 0);
+		$self->_try('I/O watcher', $io->{cb}, 0);
 	} elsif ($mode == POE_IO_WRITE) {
-		$self->_sandbox('Write', $io->{cb}, 1);
+		$self->_try('I/O watcher', $io->{cb}, 1);
 	} else {
 		die "Unknown POE I/O mode $mode";
 	}
@@ -355,18 +342,26 @@ Mojo::Reactor::POE - POE backend for Mojo::Reactor
 
   # Watch if handle becomes readable or writable
   my $reactor = Mojo::Reactor::POE->new;
-  $reactor->io($handle => sub {
+  $reactor->io($first => sub {
     my ($reactor, $writable) = @_;
-    say $writable ? 'Handle is writable' : 'Handle is readable';
+    say $writable ? 'First handle is writable' : 'First handle is readable';
   });
 
   # Change to watching only if handle becomes writable
-  $reactor->watch($handle, 0, 1);
+  $reactor->watch($first, 0, 1);
+
+  # Turn file descriptor into handle and watch if it becomes readable
+  my $second = IO::Handle->new_from_fd($fd, 'r');
+  $reactor->io($second => sub {
+    my ($reactor, $writable) = @_;
+    say $writable ? 'Second handle is writable' : 'Second handle is readable';
+  })->watch($second, 1, 0);
 
   # Add a timer
   $reactor->timer(15 => sub {
     my $reactor = shift;
-    $reactor->remove($handle);
+    $reactor->remove($first);
+    $reactor->remove($second);
     say 'Timeout!';
   });
 
@@ -421,6 +416,12 @@ Restart timer. Note that this method requires an active timer.
 Watch handle for I/O events, invoking the callback whenever handle becomes
 readable or writable.
 
+  # Callback will be invoked twice if handle becomes readable and writable
+  $reactor->io($handle => sub {
+    my ($reactor, $writable) = @_;
+    say $writable ? 'Handle is writable' : 'Handle is readable';
+  });
+
 =head2 is_running
 
   my $bool = $reactor->is_running;
@@ -433,6 +434,13 @@ Check if reactor is running.
 
 Construct a new L<Mojo::Reactor::POE> object.
 
+=head2 next_tick
+
+  my $undef = $reactor->next_tick(sub {...});
+
+Invoke callback as soon as possible, but not before returning or other
+callbacks that have been registered with this method, always returns C<undef>.
+
 =head2 one_tick
 
   $reactor->one_tick;
@@ -440,6 +448,11 @@ Construct a new L<Mojo::Reactor::POE> object.
 Run reactor until an event occurs or no events are being watched anymore. Note
 that this method can recurse back into the reactor, so you need to be careful.
 
+  # Don't block longer than 0.5 seconds
+  my $id = $reactor->timer(0.5 => sub {});
+  $reactor->one_tick;
+  $reactor->remove($id);
+
 =head2 recurring
 
   my $id = $reactor->recurring(0.25 => sub {...});
@@ -467,6 +480,9 @@ Remove all handles and timers.
 Start watching for I/O and timer events, this will block until L</"stop"> is
 called or no events are being watched anymore.
 
+  # Start reactor only if it is not running already
+  $reactor->start unless $reactor->is_running;
+
 =head2 stop
 
   $reactor->stop;
@@ -487,6 +503,18 @@ seconds.
 Change I/O events to watch handle for with true and false values. Note that
 this method requires an active I/O watcher.
 
+  # Watch only for readable events
+  $reactor->watch($handle, 1, 0);
+
+  # Watch only for writable events
+  $reactor->watch($handle, 0, 1);
+
+  # Watch for readable and writable events
+  $reactor->watch($handle, 1, 1);
+
+  # Pause watching for events
+  $reactor->watch($handle, 0, 0);
+
 =head1 CAVEATS
 
 If you set a timer or I/O watcher, and don't call L</"start"> or
